
(* Option *)

let () =
  Clflags.nopervasives := true;
  Clflags.no_std_include := true

let machine = ref true
let print_lambda = ref false

(** *)

open Misc
open Config
open Format
open Typedtree
open Compenv
open Lambda

let print_if ppf flag printer arg =
  if !flag then fprintf ppf "%a@." printer arg;
  arg

let (++) x f = f x

let build_res_prim =
   { Primitive.
     prim_name  = "build_res";
     prim_arity = 1;
     prim_alloc = false;
     prim_native_name = "";
     prim_native_float = false }

let transform lam =
  let rec descend = function
    | Llet (k,i,l, Lprim (Pmakeblock (_,_) , _ )) -> Llet (k,i,l, Lvar i)
    | Llet (k,i,l,l') -> Llet (k,i,l,descend l')
    | Lletrec (l, lam) -> Lletrec (l,descend lam)
    | Lsequence (l, Lprim (Pmakeblock (_,_) , _ )) -> l
    | Lsequence (l1, l2) -> Lsequence(l1, descend l2)
    | _ -> assert false
  in
  match lam with
  | Lprim (Psetglobal _, [lam]) ->
      if !machine then
        let arg1 = Ident.create "arg1" in
        let arg2 = Ident.create "arg2" in
        Lfunction (Curried, [arg1;arg2],
                   Lprim(Pccall build_res_prim,
                         [Lapply (descend lam, [Lvar arg1;Lvar arg2], Location.none)]))
      else
        Lfunction (Curried, [Ident.create "unit";Ident.create "unit"],
                   descend lam)
  | _ -> assert false

let implementation ppf sourcefile outputprefix =
  Location.input_name := sourcefile;
  Compmisc.init_path false;
  let modulename =
    String.capitalize(Filename.basename(chop_extension_if_any sourcefile)) in
  check_unit_name ppf sourcefile modulename;
  Env.set_unit_name modulename;
  let inputfile = Pparse.preprocess sourcefile in
  let env = Compmisc.initial_env() in
  try
    let gcc =
      Pparse.file ppf inputfile Parse.implementation ast_impl_magic_number
      ++ print_if ppf Clflags.dump_parsetree Printast.implementation
      ++ print_if ppf Clflags.dump_source Pprintast.structure
      ++ Typemod.type_implementation sourcefile outputprefix modulename env
      ++ print_if ppf Clflags.dump_typedtree
        Printtyped.implementation_with_coercion
      ++ Translmod.transl_implementation modulename
      ++ print_if ppf Clflags.dump_rawlambda Printlambda.lambda
      ++ Simplif.simplify_lambda
      ++ transform
      ++ print_if ppf print_lambda Printlambda.lambda
      ++ GccGccgen.compile !machine in
    Warnings.check_fatal ();
    Pparse.remove_preprocessed inputfile;
    Stypes.dump (Some (outputprefix ^ ".annot"));
    gcc
  with x ->
    Pparse.remove_preprocessed inputfile;
    Stypes.dump (Some (outputprefix ^ ".annot"));
    raise x

let build_gcc ppf outputprefix files =
  let _ = Sys.command "rm -f /tmp/big_file.ml" in
  List.iter
    (fun s ->
       ignore (
         ignore(Sys.command
                  (Printf.sprintf "echo \"# 1 \\\"%s\\\"\" >> /tmp/big_file.ml" s));
         Sys.command (Printf.sprintf "cat %s >> /tmp/big_file.ml" s)
       ) )
    files;
  implementation ppf "/tmp/big_file.ml" outputprefix
