(* ce fichier est nommé ainsi en honneur à ocamlbuild *)

type ghc =
    { data : int array ;
      code : instr array ;
      mutable pc : int ;
      mutable register0 : int;
      mutable register1 : int;
      mutable register2 : int;
      mutable register3 : int;
      mutable register4 : int;
      mutable register5 : int;
      mutable register6 : int;
      mutable register7 : int;
    }

let create () =
  let data = array_create 0 in
  let code = array_create Undefined in
  let pc = 0 in
  let register0 = 0 in
  let register1 = 0 in
  let register2 = 0 in
  let register3 = 0 in
  let register4 = 0 in
  let register5 = 0 in
  let register6 = 0 in
  let register7 = 0 in
  {
    data ; code ; pc ;
    register0; register1 ; register2 ;
    register3 ; register4 ; register5 ;
    register6 ; register7 ;
  }

let interp ghc world index =
  let wghostindex = match list_nth world.ghosts index with
      Some x -> x | None -> assert false in
  let initialdir = wghostindex.direction in
  let dir = ref initialdir in
  let rec interp i =
    if i = 1024 then ()
    else match aget ghc.code (ghc.pc) with
      | Halt -> ()
      | instr ->
          let pc = ghc.pc in
          interp_instr instr ;
          if ghc.pc <> pc then ghc.pc <- ghc.pc + 1 ;
          interp (i + 1)
  and interp_instr = function
    | Jmp src ->
        ghc.pc <- interp_rvalue src
    | Set (dst, src) ->
        interp_set_lvalue dst (interp_rvalue src)
    | Arith (op, dst, src) ->
        interp_set_lvalue dst (interp_arith_op op (interp_lvalue dst) (interp_rvalue src))
    | Cmp (op, addr, x, y) ->
        if interp_cmp_op op (interp_rvalue x) (interp_rvalue y) then
          ghc.pc <- addr
    | Interrupt 0 ->
        dir := match ghc.register0 with
          | 0 -> Up
          | 1 -> Right
          | 2 -> Down
          | 3 -> Left
          | _ -> initial_direction
    | Interrupt 1 ->
        let { x ; y } = world.lambda_man.location in
        ghc.register0 <- x ;
        ghc.register1 <- y
    | Interrupt 2 -> () (* no lambda man 2 -> skip *)
    | Interrupt 3 ->
        ghc.register0 <- index ;
    | Interrupt 4 ->
        let index = ghc.register0 in
        let { x ; y } = wghostindex.starting_location in
        ghc.register0 <- x ;
        ghc.register1 <- y
    | Interrupt 5 ->
        let index = ghc.register0 in
        let { x ; y } = wghostindex.current_location in
        ghc.register0 <- x ;
        ghc.register1 <- y
    | Interrupt 6 ->
        let index = ghc.register0 in
        let { vitality ; direction } = wghostindex in
        ghc.register0 <- encode_vitality vitality ;
        ghc.register1 <- encode_direction direction
    | Interrupt 7 ->
        let x = ghc.register0 in
        let y = ghc.register1 in
        ghc.register0 <- encode_cell world.map.(y).(x)
    | Interrupt 8 ->
        trace ghc
    | Interrupt i ->
        failwith (Printf.sprintf "unknown interrupt %d" i)
    | Halt ->
        assert false
    | Undefined ->
        failwith "jumped out of the GHC program"
  and interp_rvalue = function
    | Reg 0 -> ghc.register0
    | Reg 1 -> ghc.register1
    | Reg 2 -> ghc.register2
    | Reg 3 -> ghc.register3
    | Reg 4 -> ghc.register4
    | Reg 5 -> ghc.register5
    | Reg 6 -> ghc.register6
    | Reg 7 -> ghc.register7
    | Cst n -> n
    | Mem_at_reg 0 -> aget ghc.data ghc.register0
    | Mem_at_reg 1 -> aget ghc.data ghc.register1
    | Mem_at_reg 2 -> aget ghc.data ghc.register2
    | Mem_at_reg 3 -> aget ghc.data ghc.register3
    | Mem_at_reg 4 -> aget ghc.data ghc.register4
    | Mem_at_reg 5 -> aget ghc.data ghc.register5
    | Mem_at_reg 6 -> aget ghc.data ghc.register6
    | Mem_at_reg 7 -> aget ghc.data ghc.register7
    | Mem_at_cst n -> aget ghc.data n
    | Pc -> ghc.pc
  and interp_lvalue = function
    | Set_reg 0 -> ghc.register0
    | Set_reg 1 -> ghc.register1
    | Set_reg 2 -> ghc.register2
    | Set_reg 3 -> ghc.register3
    | Set_reg 4 -> ghc.register4
    | Set_reg 5 -> ghc.register5
    | Set_reg 6 -> ghc.register6
    | Set_reg 7 -> ghc.register7
    | Set_mem_at_reg 0 -> aget ghc.data ghc.register0
    | Set_mem_at_reg 1 -> aget ghc.data ghc.register1
    | Set_mem_at_reg 2 -> aget ghc.data ghc.register2
    | Set_mem_at_reg 3 -> aget ghc.data ghc.register3
    | Set_mem_at_reg 4 -> aget ghc.data ghc.register4
    | Set_mem_at_reg 5 -> aget ghc.data ghc.register5
    | Set_mem_at_reg 6 -> aget ghc.data ghc.register6
    | Set_mem_at_reg 7 -> aget ghc.data ghc.register7
    | Set_mem_at_cst n -> aget ghc.data n
  and interp_set_lvalue dst src = match dst with
    | Set_reg 0 -> ghc.register0 <- src
    | Set_reg 1 -> ghc.register1 <- src
    | Set_reg 2 -> ghc.register2 <- src
    | Set_reg 3 -> ghc.register3 <- src
    | Set_reg 4 -> ghc.register4 <- src
    | Set_reg 5 -> ghc.register5 <- src
    | Set_reg 6 -> ghc.register6 <- src
    | Set_reg 7 -> ghc.register7 <- src
    | Set_mem_at_reg 0 -> aset ghc.data ghc.register0 src
    | Set_mem_at_reg 1 -> aset ghc.data ghc.register1 src
    | Set_mem_at_reg 2 -> aset ghc.data ghc.register2 src
    | Set_mem_at_reg 3 -> aset ghc.data ghc.register3 src
    | Set_mem_at_reg 4 -> aset ghc.data ghc.register4 src
    | Set_mem_at_reg 5 -> aset ghc.data ghc.register5 src
    | Set_mem_at_reg 6 -> aset ghc.data ghc.register6 src
    | Set_mem_at_reg 7 -> aset ghc.data ghc.register7 src
    | Set_mem_at_cst n -> aset ghc.data n src
  and interp_arith_op op x y = match op with
    | Add -> modulo (x + y) 256
    | Sub -> modulo (x - y) 256
    | Mul -> modulo (x * y) 256
    | Div -> modulo (x / y) 256
    | And -> x land y
    | Or -> x lor y
    | Xor -> (x lxor y)
  and interp_cmp_op op x y = match op with
    | Lt -> x < y
    | Eq -> x = y
    | Gt -> x > y
  in
  ghc.pc <- 0 ;
  interp 0 ;
  !dir
