
open GccTypes

let trace = ref false
let one = ref false

let arg_list = Arg.align [
    "-dlambda", Arg.Set GccComp.print_lambda,
    "print intermediate lambda";
    "-trace", Arg.Set trace,
    "trace every instruction";
    "-one", Arg.Set one,
    "files of the data stack is not redudex to 'Int 1' after a successful run";
  ]

let files_to_concat = ref []

let arg_anon s =
  try
    if Filename.check_suffix s ".ml" then
      files_to_concat := s :: !files_to_concat
    else begin
      Printf.eprintf "Error: don't know what to do with %S\n%!" s;
      exit 2
    end
  with x ->
    Errors.report_error Format.err_formatter x;
    exit 2

let arg_usage =
  Printf.sprintf "%s [options] file1.ml [file2.ml...]" Sys.argv.(0)

let _ = Arg.parse arg_list arg_anon arg_usage


let () =
  if List.length !files_to_concat = 0 then begin
    prerr_endline "Usage: ./test_gcc file1.ml [file2.ml...]";
    exit 2
  end


let build () =
  try
    GccComp.machine := false;
    let files = List.rev !files_to_concat in
    let outputprefix = Compenv.output_prefix (List.hd !files_to_concat) in
    GccComp.build_gcc Format.err_formatter outputprefix files
  with exn ->
    Errors.report_error Format.err_formatter exn;
    exit 1


let run gcc =
  let state = GccRun.init_state gcc in
  let trace =
    if !trace
    then Some (GccRun.simple_trace Format.err_formatter)
    else None
  in
  let debug = GccRun.simple_debug Format.err_formatter in
  let runner = GccRun.arg_runner ~debug ?trace gcc in
  try
    let resulting_state = GccRun.instrs_runner runner state in
    List.iter
      (Format.eprintf "%a@." (GccRun.print_value runner))
      resulting_state.data;
      if !one && resulting_state.data <> [TAG_INT 1] then exit 1
  with FAULT fault ->
    match fault with
    | TAG_MISMATCH s ->
        Format.eprintf "FAULT: TAG_MISMATCH %s@." s;
        exit 3
    | CONTROL_MISMATCH s ->
        Format.eprintf "FAULT: CONTROL_MISMATCH %s@." s;
        exit 3
    | FRAME_MISMATCH (i1, i2) ->
        Format.eprintf "FAULT: FRAME_MISMATCH %d %d@." i1 i2;
        exit 3
    | EMPTY_STACK s ->
        Format.eprintf "FAULT: EMPTY_STACK %s@." s;
        exit 3
    | EMPTY_FRAME s ->
        Format.eprintf "FAULT: EMPTY_FRAME %s@." s;
        exit 3

let () = run (build ())

