open Ghc
open Ghc_lang

let up = 0
let right = 1
let down = 2
let left = 3

let goto_lambda1 (flip_activation:bool) : code =
  let ghost_x = var 0 in
  let ghost_y = var 1 in
  let lambda_x = var 2 in
  let lambda_y = var 3 in
  let dx = var 4 in
  let dy = var 5 in

  let vitality = var 6 in
  let curr_direction = var 7 in
  let flip = var 8 in
  let truc_flip_activated = if flip_activation then 1 else 3 in
  let iter_ghost = var 9 in
  let ghost_index = var 10 in

  let other_ghost_x = var 11 in
  let other_ghost_y = var 12 in
  let flip' = var 13 in
  let place_info = var 14 in
  let some_change = var 15 in

  let dir = var 100 in

  let seen_pos = var 101 in
  let seen_scratch_pos = var 102 in
  let n_seen = 16 in (* must be power of 2 *)
  let seen_x_table, seen_y_table =
    let n = 120 in
    var n, var (n + n_seen) in

  let expected_ghosts = 32 in

  [

    "entry",
    [
      Lambda_man_pos (Var lambda_x, Var lambda_y, 1);
      Ghost_info (Ghost_current_pos, Var ghost_x,
                  Var ghost_y, Ghost_index);
      Ghost_info (Ghost_vit_dir,
                  Var vitality,
                  Var curr_direction,
                  Ghost_index);
      Var ghost_index =^ Ghost_index;
      Var flip =^ ((Var flip +^ Const 1) &&^ Const 1);
      Var flip' =^ ((Var flip' +^ Const 1) &&^ Const 3);
      Var some_change =^ Const 0;
    ],
    Cond (Lt, "right", Var ghost_x, Var lambda_x);

    "left",
    [
      Var dx =^ Var ghost_x -^ Var lambda_x;
    ],
    Cond2 (Gt, "leftup", "leftdown", Var ghost_y, Var lambda_y);

    "right",
    [
      Var dx =^ Var lambda_x -^ Var ghost_x;
    ],
    Cond2 (Gt, "rightup", "rightdown", Var ghost_y, Var lambda_y);

    "leftdown",
    [
      Var dy =^ Var lambda_y -^ Var ghost_y;
    ],
    Cond2 (Gt, "godown", "goleft", Var dy, Var dx);

    "leftup",
    [
      Var dy =^ Var ghost_y -^ Var lambda_y;
    ],
    Cond2 (Gt, "goup", "goleft", Var dy, Var dx);

    "rightdown",
    [
      Var dy =^ Var lambda_y -^ Var ghost_y;
    ],
    Cond2 (Gt, "godown", "goright", Var dy, Var dx);

    "rightup",
    [
      Var dy =^ Var ghost_y -^ Var lambda_y;
    ],
    Cond2 (Gt, "goup", "goright", Var dy, Var dx);




    "goleft",
    [
      Var dir =^ Const left;
      Var ghost_x =^ Var ghost_x -^ Const 1;
    ],
    Goto "checkseen";

    "godown",
    [
      Var dir =^ Const down;
      Var ghost_y =^ Var ghost_y +^ Const 1;
    ],
    Goto "checkseen";

    "goright",
    [
      Var dir =^ Const right;
      Var ghost_x =^ Var ghost_x +^ Const 1;
    ],
    Goto "checkseen";

    "goup",
    [
      Var dir =^ Const up;
      Var ghost_y =^ Var ghost_y -^ Const 1;
    ],
    Goto "checkseen";

    (* check if we have seen that position *)

    "checkseen",
    [
      Var seen_scratch_pos =^ Const 0;
    ],
    No_branch;

    "checkseen_loop",
    [
    ],
    No_branch;

    "checkseen_x",
    [
    ],
    Cond2 (Eq, "checkseen_y", "not_seen",
           Var ghost_x, Field(seen_x_table, Var seen_scratch_pos));

    "checkseen_y",
    [
      (* Debugv (Var seen_scratch_pos); *)
    ],
    Cond2 (Eq, "seen", "not_seen",
           Var ghost_y, Field(seen_y_table, Var seen_scratch_pos));

    "not_seen",
    [
      Var seen_scratch_pos =^ Var seen_scratch_pos +^ Const 1;
    ],
    Cond2 (Eq, "never_seen", "checkseen_loop", Var seen_scratch_pos, Const n_seen);

    "never_seen",
    [
      (* Debugv (Const 0); *)
      (* Debugv (Var seen_scratch_pos); *)
    ],
    Goto "exit";

(*
    "seen",
    [
      Var dir =^ ((Var dir +^ Const 2) &&^ Const 3);
      (* Debugv (Var dir); *)
      (* Debugv (Field(seen_x_table, Var seen_scratch_pos)); *)
      (* Debugv (Var seen_scratch_pos); *)
      (* Debug; *)
    ],
    No_branch;
*)

    "seen",
    [

      (* Debugv (Const 123); *)
      (* Debugv (Var dir); *)
      (* Debugv (Var curr_direction); *)

    ],
    Cond2 (Eq, "sidestep_seen", "invert_seen", Var curr_direction, Var dir); (* if going opposite *)

    "sidestep_seen",
    [
      (* Debugv (Const 0); *)
      Var dir =^ ((Var dir +^ Const truc_flip_activated +^ (Var flip *^ Const 2)) &&^ Const 3);
      Var some_change =^ Const 1;
    ],
    Goto "exit";

    "invert_seen",
    [
      (* Debugv (Const 1); *)
      Var dir =^ ((Var dir +^ Const 2) &&^ Const 3);
      Var some_change =^ Const 1;
    ],
    Goto "exit";

    "exit",
    [
      (* Debugv (Const 111); *)
      (* Debugv (Var vitality); *)
    ],
    Cond2 (Eq, "invert", "last_exit", Var vitality, Const 1); (* if fright *)

    "invert",
    [
      (* Debugv (Const 222); *)
    ],
    Cond2 (Eq, "invert_really", "sidestep", Var curr_direction, Var dir); (* if going opposite *)

    "sidestep",
    [
      Var dir =^ ((Var dir +^ Const 1) &&^ Const 3);
      Var some_change =^ Const 1;
    ],
    Goto "exit";

    "invert_really",
    [
      Var dir =^ ((Var dir +^ Const 2) &&^ Const 3);
      Var some_change =^ Const 1;
    ],
    Goto "exit";

    "last_exit",
    [
      Set_dir (Var dir); (* write in case we take to much time after... *)
      Field(seen_x_table, Var seen_pos) =^ Var ghost_x;
      Field(seen_y_table, Var seen_pos) =^ Var ghost_y;
      Var seen_pos =^ ((Var seen_pos +^ Const 1) &&^ Const (n_seen - 1));


      Var iter_ghost =^ Const 0;
    ],
    Goto "iter_ghost_loop";

    "iter_ghost_loop", [],
    Cond (Eq, "really_last_exit", Var iter_ghost, Const expected_ghosts);

    "ghost_check",
    [
      Ghost_info (Ghost_current_pos,
                  Var other_ghost_x,
                  Var other_ghost_y,
                  Var iter_ghost);
    ],
    Cond (Eq, "ghost_check_y", Var other_ghost_x, Var ghost_x);

    "ghost_end_loop",
    [
      Var iter_ghost =^ Var iter_ghost +^ Const 1;
    ],
    Goto "iter_ghost_loop";

    "ghost_check_y", [],
    Cond2 (Eq, "ghost_equal", "ghost_end_loop", Var other_ghost_y, Var ghost_y);

    "ghost_equal",
    [
      (* Debugv (Const 222); *)
      (* Debugv (Var dir); *)
      Var dir =^ ((Var dir +^ Const truc_flip_activated +^ (Var flip' *^ Const 2)) &&^ Const 3);
      Var some_change =^ Const 1;
      Set_dir (Var dir);
      (* Debugv (Var dir); *)
    ],
    Goto "really_last_exit";

    "really_last_exit",
    [],
    Cond2 (Eq, "wall_turner", "really_last_exit2",
           Var some_change, Const 0);

    "wall_turner",
    [
      Var place_info =^ Place(Var ghost_x, Var ghost_y);
    ],
    Cond2 (Eq, "handle_wall", "really_last_exit2",
           Var place_info, Const 0);

    "handle_wall",
    [
      Var dir =^ ((Var dir +^ Const truc_flip_activated +^ (Var flip *^ Const 2)) &&^ Const 3);
    ],
    No_branch;

    "really_last_exit2",
    [
      (* Debugv (Var dx); *)
      (* Debugv (Var dy); *)
      (* Debugv (Var dir); *)

      (* Debugv (Var seen_pos); *)
      (* Debugv (Field(seen_x_table, Var seen_pos)); *)
      (* Debugv (Field(seen_y_table, Var seen_pos)); *)


      (* Debugv (Var seen_pos); *)
      (* Debugv (Field(seen_x_table, Var seen_pos)); *)
      (* Debugv (Field(seen_y_table, Var seen_pos)); *)

    ],
    End;
  ]

(*
let goto_lambda1 : code =
  let ghost_x = var 0 in
  let ghost_y = var 1 in
  let lambda_x = var 2 in
  let lambda_y = var 3 in
  let dx = var 4 in
  let dy = var 5 in
  let dir = var 100 in

  let seen_pos = var 101 in
  let seen_scratch_pos = var 102 in
  let n_seen = 16 in (* must be power of 2 *)
  let seen_x_table, seen_y_table =
    let n = 120 in
    var n, var (n + n_seen) in

  [

    (* "entry", *)
    (* [ *)
    (*   Lambda_man_pos (Var lambda_x, Var lambda_y, 1); *)
    (*   Ghost_info (Ghost_current_pos, Var ghost_x, *)
    (*               Var ghost_y, Ghost_index); *)
    (* ], *)
    (* No_branch; *)
    (* Cond (Lt, "right", Var ghost_x, Var lambda_x); *)

    "left",
    [
      Var ghost_x =^ Const 22;
      Var lambda_x =^ Const 33;
      Var dx =^ Var ghost_x -^ Var lambda_x;
      Debugv (Const 0);
      Debugv (Var ghost_x);
      Debugv (Var lambda_x);
      Debugv (Var dx);
    ],
    End;

  ]
*)

let goto_lambda1' = build_blocks (goto_lambda1 true)

let ghc =
  Format.eprintf "length %i@." (List.length goto_lambda1');
  { data = [||] ;
    code = Array.of_list goto_lambda1';
    pc = 0 ;
    registers = [||] }

let output_code f ghc =
  let oc = open_out f in
  let ppf = Format.formatter_of_out_channel oc in
  Ghc.print ppf ghc;
  Format.fprintf ppf "@.";
  close_out oc

let mk_ghc code =
  Format.eprintf "length %i@." (List.length code);
  { data = [||] ;
    code = Array.of_list code;
    pc = 0 ;
    registers = [||] }

let mk_code1 code_of_flip = connardize_zerodiv (code_of_flip true)
let mk_code2 code_of_flip = connardize_goto255 (code_of_flip false)
let mk_code3 = mk_code1

let mk_3_ghosts cof =
  List.iter
    (fun (f, c) -> output_code f (mk_ghc (build_blocks c)))
    [("ghost0.ghc", mk_code1 cof);
     ("ghost1.ghc", mk_code2 cof);
     ("ghost2.ghc", mk_code3 cof)]
  

let () =
  mk_3_ghosts goto_lambda1
  (* let oc = open_out "follower.ghc" in *)
  (* let ppf = Format.formatter_of_out_channel oc in *)
  (* Ghc.print ppf ghc; *)
  (* Format.fprintf ppf "@."; *)
  (* close_out oc *)

