type instr =
  | Jmp of rvalue
  | Set of lvalue * rvalue
  | Arith of arith_op * lvalue * rvalue
  | Cmp of cmp_op * int * rvalue * rvalue
  | Interrupt of int
  | Halt
  | Undefined
 and rvalue =
   | Reg of reg
   | Cst of int
   | Mem_at_reg of reg
   | Mem_at_cst of int
   | Pc
 and lvalue =
   | Set_reg of reg
   | Set_mem_at_reg of reg
   | Set_mem_at_cst of int
 and arith_op =
   | Add | Sub | Mul | Div
   | And | Or | Xor
 and cmp_op =
   | Lt | Eq  | Gt
 and reg = int
 and interrupt = int

type ghc =
    { data : int array ;
      code : instr array ;
      mutable pc : int ;
      registers : int array }

val from_file : string -> ghc
val print : Format.formatter -> ghc -> unit
val interp : ghc -> World.world -> int -> World.direction
